//this file is part of Notepad++
//Copyright (C)2007 Don HO <donho@altern.org>
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "PluginInterface.h"
#include "NppTools.h"

const char PLUGIN_NAME[] = "Base64";
const int nbFunc = 3;

NppData nppData;
FuncItem funcItem[nbFunc];

void convertToBase64FromAscii();
void convertToBase64FromAscii_B64Format();
void convertToAsciiFromBase64();

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  reasonForCall, 
                       LPVOID lpReserved )
{
	switch (reasonForCall)
	{
		case DLL_PROCESS_ATTACH:
		{
			funcItem[0]._pFunc = convertToBase64FromAscii;
			funcItem[1]._pFunc = convertToBase64FromAscii_B64Format;
			funcItem[2]._pFunc = convertToAsciiFromBase64;

			strcpy(funcItem[0]._itemName, "Base64 Encode");
			strcpy(funcItem[1]._itemName, "Base64 Encode with Unix EOL");
			strcpy(funcItem[2]._itemName, "Base64 Decode");

			funcItem[0]._init2Check = false;
			funcItem[1]._init2Check = false;
			funcItem[2]._init2Check = false;

			// If you don't need the shortcut, you have to make it NULL
			funcItem[0]._pShKey = NULL;
			funcItem[1]._pShKey = NULL;
			funcItem[2]._pShKey = NULL;
		}
		break;

		case DLL_PROCESS_DETACH:
		break;

		case DLL_THREAD_ATTACH:
		break;

		case DLL_THREAD_DETACH:
		break;
	}
	return TRUE;
}

extern "C" __declspec(dllexport) void setInfo(NppData notpadPlusData)
{
	nppData = notpadPlusData;
}

extern "C" __declspec(dllexport) const char * getName()
{
	return PLUGIN_NAME;
}

extern "C" __declspec(dllexport) FuncItem * getFuncsArray(int *nbF)
{
	*nbF = nbFunc;
	return funcItem;
}

extern "C" __declspec(dllexport) void beNotified(SCNotification *notifyCode)
{	
}

// Here you can process the Npp Messages 
// I will make the messages accessible little by little, according to the need of plugin development.
// Please let me know if you need to access to some messages :
// http://sourceforge.net/forum/forum.php?forum_id=482781
//
extern "C" __declspec(dllexport) LRESULT messageProc(UINT Message, WPARAM wParam, LPARAM lParam)
{
	return TRUE;
}

HWND getCurrentScintillaHandle() {
    int currentEdit;
    ::SendMessage(nppData._nppHandle, NPPM_GETCURRENTSCINTILLA, 0, (LPARAM)&currentEdit);
	return (currentEdit == 0)?nppData._scintillaMainHandle:nppData._scintillaSecondHandle;
};

size_t asciiToBase64(char *dest, const unsigned char *asciiStr, size_t asciiStrLen)
{
	//size_t asciiStrLen = strlen(asciiStr);
	size_t nbLoop = asciiStrLen / 3;
	size_t nbByteRemain = asciiStrLen % 3;

	size_t i = 0, k = 0;
	for ( ; i < nbLoop ; i++)
	{
		size_t j = i * 3;
		unsigned char p1, p2;

		unsigned char a = asciiStr[j] >> 2;
		dest[k++] = base64Array[a];
		p1 = asciiStr[j] << 6;
		p2 = asciiStr[j+1] >> 2;
		unsigned char b = p1 | p2;
		b >>= 2;
		dest[k++] = base64Array[b];

		p1 = asciiStr[j+1] << 4;
		p2 = asciiStr[j+2] >> 4;
		unsigned char c = p1 | p2;
		c >>= 2;
		dest[k++] = base64Array[c];

		p2 = asciiStr[j+2] << 2;
		unsigned char d = p2 >> 2;
		dest[k++] = base64Array[d];
	}

	if (nbByteRemain == 1)
	{
		size_t j = i * 3;
		unsigned char p;
		dest[k++] = base64Array[asciiStr[j] >> 2];

		p = asciiStr[j] << 6;
		dest[k++] = base64Array[p >> 2];

		dest[k++] = '=';
		dest[k++] = '=';
	}
	else if (nbByteRemain == 2)
	{
		size_t j = i * 3;
		unsigned char p1, p2;

		dest[k++] = base64Array[asciiStr[j] >> 2];

		p1 = asciiStr[j] << 6;
		p2 = asciiStr[j+1] >> 2;
		unsigned char b = p1 | p2;
		dest[k++] = base64Array[b >> 2];

		p1 = asciiStr[j+1] << 4;
		dest[k++] = base64Array[p1 >> 2];

		dest[k++] = '=';
	}

	dest[k] = '\0';
	return k;//
}

int base64ToAscii(char *dest, const char *base64Str)
{
	size_t b64StrLen = strlen(base64Str);	
	size_t nbLoop = b64StrLen / 4;
	//if (b64StrLen % 4)
	//	return false;
	size_t i = 0;
	int k = 0;

	enum {b64_just, b64_1padded, b64_2padded} padd = b64_just;
	for ( ; i < nbLoop ; i++)
	{
		size_t j = i * 4;
		unsigned char uc0, uc1, uc2, uc3, p0, p1;

		uc0 = base64IndexArray[base64Str[j]];
		uc1 = base64IndexArray[base64Str[j+1]];
		uc2 = base64IndexArray[base64Str[j+2]];
		uc3 = base64IndexArray[base64Str[j+3]];

		if ((uc0 == -1) || (uc1 == -1) || (uc2 == -1) || (uc3 == -1))
			return -1;

		if (base64Str[j+2] == '=') // && (uc3 == '=')
		{
			uc2 = uc3 = 0;
			padd = b64_2padded;
		}
		else if (base64Str[j+3] == '=')
		{
			uc3 = 0;
			padd = b64_1padded;
		}
		p0 = uc0 << 2;
		p1 = uc1 << 2;
		p1 >>= 6;
		dest[k++] = p0 | p1;

		p0 = uc1 << 4;
		p1 = uc2 << 2;
		p1 >>= 4;
		dest[k++] = p0 | p1;

		p0 = uc2 << 6;
		p1 = uc3;
		dest[k++] = p0 | p1;
	}

	//dest[k] = '\0';
	if (padd == b64_1padded)
	//	dest[k-1] = '\0';
		return k-1;
	else if (padd == b64_2padded)
	//	dest[k-2] = '\0';
		return k-2;

	return k;
}

void convertToBase64FromAscii()
{
	asciiToBase64(false);
}

void convertToBase64FromAscii_B64Format()
{
	asciiToBase64(true);
}

void asciiToBase64(bool doFormat)
{
	HWND hCurrScintilla = getCurrentScintillaHandle();

    size_t start = ::SendMessage(hCurrScintilla, SCI_GETSELECTIONSTART, 0, 0);
	size_t end = ::SendMessage(hCurrScintilla, SCI_GETSELECTIONEND, 0, 0);
	if (end < start)
	{
		size_t tmp = start;
		start = end;
		end = tmp;
	}

	size_t asciiTextLen = end - start;
	//size_t asciiTextLen = ::SendMessage(hCurrScintilla, SCI_GETSELTEXT, 0, 0);
    if (asciiTextLen == 0) return;

    unsigned char * pAsciiText = new unsigned char[asciiTextLen+1];
    
    ::SendMessage(hCurrScintilla, SCI_GETSELTEXT, 0, (LPARAM)pAsciiText);

	size_t b64Len = getBase64LenFromAsciiLen(asciiTextLen);
	char * pBase64Text = new char[b64Len+1];

	asciiToBase64(pBase64Text, pAsciiText, asciiTextLen);
	
	char *pBase64Text2Display = pBase64Text;
	if (doFormat)
	{
		size_t nbEOL = b64Len/64; //nb 0x0A
		pBase64Text2Display = new char[b64Len+nbEOL+1];

		size_t m = 0, n = 0;
		for ( ; m < b64Len ; m++)
		{
			pBase64Text2Display[n++] = pBase64Text[m];

			if ((m+1)%64 == 0)
				pBase64Text2Display[n++] = 0x0A;
		}
		pBase64Text2Display[n] = '\0';
	}

	::SendMessage(hCurrScintilla, SCI_REPLACESEL, 0, (LPARAM)pBase64Text2Display);
	::SendMessage(hCurrScintilla, SCI_SETSEL, start, start+strlen(pBase64Text2Display));

	delete [] pAsciiText;
	delete [] pBase64Text;

	if (doFormat)
		delete [] pBase64Text2Display;
}

void convertToAsciiFromBase64()
{
	HWND hCurrScintilla = getCurrentScintillaHandle();
    size_t textLength = ::SendMessage(hCurrScintilla, SCI_GETSELTEXT, 0, 0);
    if (textLength == 0) return;

	char * selectedText = new char[textLength];
    char * pBase64Text = new char[textLength];
    ::SendMessage(hCurrScintilla, SCI_GETSELTEXT, 0, (LPARAM)selectedText);

	size_t m = 0, n = 0;
	for ( ; m < textLength ; )
	{
		if (selectedText[m] == 0x0A || selectedText[m] == 0x0D)
			m++;
		else
			pBase64Text[n++] = selectedText[m++];
	}

	size_t asciiLen = getAsciiLenFromBase64Len(n-1);
	if (!asciiLen)
	{
		::MessageBox(nppData._nppHandle, "Length of selected text (not including EOL) to be decoded is invalid.\rIt should be mod 4.", "Base64", MB_OK);
		return;
	}

	char * pAsciiText = new char[asciiLen + 1];

	int len = base64ToAscii(pAsciiText, pBase64Text);

	if (len == -1)
		::MessageBox(nppData._nppHandle, "Problem!", "Base64", MB_OK);
	else
	{
		size_t start = ::SendMessage(hCurrScintilla, SCI_GETSELECTIONSTART, 0, 0);
		size_t end = ::SendMessage(hCurrScintilla, SCI_GETSELECTIONEND, 0, 0);
		if (end < start)
		{
			size_t tmp = start;
			start = end;
			end = tmp;
		}
		::SendMessage(hCurrScintilla, SCI_SETTARGETSTART, start, 0);
		::SendMessage(hCurrScintilla, SCI_SETTARGETEND, end, 0);
		::SendMessage(hCurrScintilla, SCI_REPLACETARGET, len, (LPARAM)pAsciiText);
		::SendMessage(hCurrScintilla, SCI_SETSEL, start, start+len);
	}	

	delete [] pAsciiText;
	delete [] pBase64Text;
	delete [] selectedText;
}

